<ul class="navbar-icon-links navbar-nav"
    aria-label="{{ theme_icon_links_label }}">
  <li class="nav-item">
    <a class="nav-link github-button" href="https://github.com/sebp/scikit-survival" rel="noopener" target="_blank" title="{{ project }}">
        <span class="github-icon"><i class="fa-brands fa-github"></i></span>
    </a>
  </li>
</ul>